@@ -1,9 +1,9 @@
 import asyncio
 import io
-import os
 import time
+from collections import defaultdict
 from datetime import datetime
-from typing import Dict, Optional, Sequence, Tuple
+from typing import Dict, List, Mapping, Optional, Sequence, Tuple
 
 import aiohttp
 import pytest
@@ -11,10 +11,15 @@ from _pytest.config import Config
 from _pytest.terminal import TerminalReporter
 from discord import AsyncWebhookAdapter, Colour, Embed, File, Webhook
 from discord.errors import Forbidden, HTTPException, InvalidArgument, NotFound
-from typepy import Bool, Integer, StrictLevel
-from typepy.error import TypeConversionError
+from pytest_md_report.plugin import extract_pytest_stats
 
-from ._const import Default, HelpMsg, Option
+from ._const import HelpMsg, Option, TestResultType
+from ._opt_retriever import DiscordOptRetriever
+
+
+MAX_EMBED_LEN = 2048
+MAX_EMBEDS_LEN = 6000
+MAX_EMBED_CT = 10
 
 
 def pytest_addoption(parser):
@@ -97,99 +102,6 @@ def pytest_addoption(parser):
     )
 
 
-class DiscordOptRetriever:
-    def __init__(self, config: Config):
-        self.__config = config
-
-    def retrieve_webhook_url(self) -> Optional[str]:
-        return self.__retrieve_discord_opt(Option.DISCORD_WEBHOOK)
-
-    def retrieve_verbosity_level(self) -> int:
-        config = self.__config
-        discord_opt = Option.DISCORD_VERBOSE
-        verbosity_level = None
-
-        if hasattr(config.option, discord_opt.inioption_str):
-            verbosity_level = getattr(config.option, discord_opt.inioption_str)
-
-        if verbosity_level is not None and verbosity_level < 0:
-            verbosity_level = None
-
-        if verbosity_level is None:
-            verbosity_level = self._to_int(os.environ.get(discord_opt.envvar_str))
-
-        if verbosity_level is None:
-            verbosity_level = self._to_int(config.getini(discord_opt.inioption_str))
-
-        if verbosity_level is None:
-            verbosity_level = config.option.verbose
-
-        return verbosity_level
-
-    def retrieve_username(self) -> str:
-        username = self.__retrieve_discord_opt(Option.DISCORD_USERNAME)  # type: ignore
-
-        if not username:
-            return Default.USERNAME
-
-        return username
-
-    def retrieve_success_icon(self) -> Optional[str]:
-        return self.__retrieve_discord_opt(Option.DISCORD_SUCCESS_ICON)
-
-    def retrieve_skip_icon(self) -> Optional[str]:
-        return self.__retrieve_discord_opt(Option.DISCORD_SKIP_ICON)
-
-    def retrieve_fail_icon(self) -> Optional[str]:
-        return self.__retrieve_discord_opt(Option.DISCORD_FAIL_ICON)
-
-    def retrieve_attach_file(self) -> bool:
-        config = self.__config
-        discord_opt = Option.DISCORD_ATTACH_FILE
-        value = None
-
-        if hasattr(config.option, discord_opt.inioption_str):
-            value = getattr(config.option, discord_opt.inioption_str)
-
-        if value is None:
-            try:
-                value = Bool(
-                    os.environ.get(discord_opt.envvar_str), strict_level=StrictLevel.MIN
-                ).convert()
-            except TypeConversionError:
-                value = None
-
-        if value is None:
-            value = config.getini(discord_opt.inioption_str)
-
-        if value is None:
-            return False
-
-        return value
-
-    def __retrieve_discord_opt(self, discord_opt: Option) -> Optional[str]:
-        config = self.__config
-        value = None
-
-        if hasattr(config.option, discord_opt.inioption_str):
-            value = getattr(config.option, discord_opt.inioption_str)
-
-        if not value:
-            value = os.environ.get(discord_opt.envvar_str)
-
-        if not value:
-            value = config.getini(discord_opt.inioption_str)
-
-        return value
-
-    @staticmethod
-    def _to_int(value) -> Optional[int]:
-        try:
-            return Integer(value, strict_level=StrictLevel.MIN).convert()
-        except TypeConversionError:
-            return None
-
-
 def _normalize_stat_name(name: str) -> str:
     if name == "error":
         return "errors"
@@ -215,7 +127,7 @@ def _decorate_code_block(lang: str, text: str) -> str:
     return "```{lang}\n{body}\n```\n".format(lang=lang, body=text)
 
 
-def _concat_longrepr(reporter: TerminalReporter) -> str:
+def _extract_longrepr(reporter: TerminalReporter) -> List[str]:
     messages = []
 
     for stat_key, values in reporter.stats.items():
@@ -233,11 +145,11 @@ def _concat_longrepr(reporter: TerminalReporter) -> str:
             except AttributeError:
                 pass
 
-    return "\n\n\n".join(messages)
+    return messages
 
 
-def make_mdreport(config: Config) -> str:
-    from pytest_md_report import ColorPoicy, ZerosRender, make_md_report, retrieve_stat_count_map
+def _make_md_report(config: Config) -> str:
+    from pytest_md_report import ColorPolicy, ZerosRender, make_md_report, retrieve_stat_count_map
 
     opt_retriever = DiscordOptRetriever(config)
     verbosity_level = opt_retriever.retrieve_verbosity_level()
@@ -257,7 +169,7 @@ def make_mdreport(config: Config) -> str:
         config.option.md_report_margin = 1
 
     try:
-        config.option.md_report_color = ColorPoicy.NEVER
+        config.option.md_report_color = ColorPolicy.NEVER
         config.option.md_report_zeros = ZerosRender.EMPTY
 
         return make_md_report(config, reporter, stat_count_map)
@@ -266,7 +178,27 @@ def make_mdreport(config: Config) -> str:
         config.option.md_report_zeros = stash_md_report_zeros
 
 
-logs = []
+_logs = []
+
+
+def extract_result_type(pytest_stats: Mapping[str, int]) -> TestResultType:
+    if sum([pytest_stats[name] for name in ("failed", "error")]):
+        return TestResultType.FAIL
+
+    if (
+        sum([pytest_stats[name] for name in ("skipped", "xfailed", "xpassed")])
+        and pytest_stats["passed"] == 0
+    ):
+        return TestResultType.SKIP
+
+    return TestResultType.SUCCESS
+
+
+_result_type_to_colour = {
+    TestResultType.SUCCESS: Colour.green(),
+    TestResultType.SKIP: Colour.gold(),
+    TestResultType.FAIL: Colour.red(),
+}
 
 
 def pytest_unconfigure(config):
@@ -277,7 +209,7 @@ def pytest_unconfigure(config):
 
     verbosity_level = opt_retriever.retrieve_verbosity_level()
     reporter = config.pluginmanager.get_plugin("terminalreporter")
-    md_report = make_mdreport(config)
+    md_report = _make_md_report(config)
 
     try:
         duration = time.time() - reporter._sessionstarttime
@@ -299,14 +231,13 @@ def pytest_unconfigure(config):
         avatar_url = opt_retriever.retrieve_success_icon()
         colour = Colour.green()
 
-    if verbosity_level >= 1:
-        description = _decorate_code_block(lang="md", text="# test results\n{}".format(md_report))
-    else:
-        description = "{} in {:.1f} seconds".format(message, duration)
-
-    header = "test summary info: {} tests".format(sum(stat_count_map.values()))
     embeds = []  # type: List[Embed]
-    embed_summary = Embed(description=description, colour=colour)
+    embeds_len_ct = 100
+    exceeds_embeds_limit = False
+
+    embed_summary = Embed(
+        description="{} in {:.1f} seconds".format(message, duration), colour=colour
+    )
     embed_summary.set_footer(
         text="start at {}".format(
             datetime.fromtimestamp(reporter._sessionstarttime).strftime("%d. %b %H:%M:%S%z")
@@ -315,16 +246,60 @@ def pytest_unconfigure(config):
     embeds.append(embed_summary)
 
     if verbosity_level >= 1:
-        embeds.append(Embed(description=_concat_longrepr(reporter), colour=colour))
+        pytest_stats = extract_pytest_stats(
+            reporter=reporter,
+            outcomes=["passed", "failed", "error", "skipped", "xfailed", "xpassed"],
+            verbosity_level=max(0, verbosity_level - 1),
+        )
+        result_lines_map = defaultdict(list)
 
+        for key, stats in pytest_stats.items():
+            result_lines_map[extract_result_type(stats)].append(
+                "`{}`: {}".format(
+                    ":".join(key),
+                    ", ".join(
+                        ["`{}` {}".format(ct, outcome) for outcome, ct in stats.items() if ct > 0]
+                    ),
+                )
+            )
+
+        for result_type, result_lines in result_lines_map.items():
+            embeds.append(
+                Embed(
+                    description="\n".join(result_lines)[:MAX_EMBED_LEN],
+                    colour=_result_type_to_colour[result_type],
+                )
+            )
+
+        description = _decorate_code_block(lang="md", text="# test results\n{}".format(md_report))
+        embed_summary = Embed(description=description, colour=colour,)
+        embeds_len_ct += len(description)
+
+        longreprs = _extract_longrepr(reporter)
+        for i, description in enumerate(longreprs):
+            description = description[-MAX_EMBED_LEN:]
+
+            if len(embeds) >= MAX_EMBED_CT or (len(description) + embeds_len_ct) > MAX_EMBEDS_LEN:
+                embeds.append(
+                    Embed(
+                        description="and other {} failed".format(len(longreprs) - i), colour=colour
+                    )
+                )
+                exceeds_embeds_limit = True
+                break
+
+            embeds.append(Embed(description=description, colour=colour))
+            embeds_len_ct += len(description)
+
+    header = "test summary info: {} tests".format(sum(stat_count_map.values()))
     attach_file = None
 
-    if opt_retriever.retrieve_attach_file():
+    if opt_retriever.retrieve_attach_file() or exceeds_embeds_limit:
         attach_file = File(
             io.BytesIO(
-                "# {}\n{}\n\n{}".format(header, md_report, _concat_longrepr(reporter)).encode(
-                    "utf8"
-                )
+                "# {}\n{}\n\n{}".format(
+                    header, md_report, "\n\n".join(_extract_longrepr(reporter))
+                ).encode("utf8")
             ),
             datetime.fromtimestamp(reporter._sessionstarttime).strftime(
                 "pytest_%Y-%m-%dT%H:%M:%S.md"
@@ -366,11 +341,11 @@ async def _send_message(
         )
 
 
-@pytest.hookimpl()  # after _pytest.runner
+@pytest.hookimpl()
 def pytest_report_teststatus(report):
     if report.longreprtext:
-        logs.append(report.longreprtext)
+        _logs.append(report.longreprtext)
     if report.capstdout:
-        logs.append(report.capstdout)
+        _logs.append(report.capstdout)
     if report.capstderr:
-        logs.append(report.capstderr)
+        _logs.append(report.capstderr)
